from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time
import json
import subprocess
import logging
from collections import namedtuple, OrderedDict

log = logging.getLogger("fbjbod_lib")

POWER_SET_WAIT = 8
FAULT_SET_WAIT = 8


class FbjbodRunError(RuntimeError):

    def __init__(self, cmd, returncode, msg):
        self.cmd = cmd
        self.returncode = returncode
        super(FbjbodRunError, self).__init__(msg)


class classproperty(property):
    """
    Allow for getter property usage on classmethods
    """
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


GPIOStatus = namedtuple("GPIOStatus", ["field_name", "description", "status"])


PhysicalErrorCounters = namedtuple("PhysicalErrorCounters",
                                   ["error_buffer_id", "invalid_dword",
                                    "running_disparity", "dword_sync_loss",
                                    "physical_reset_problems"])


Fan = namedtuple("Fan", ["name", "rpms"])


LEDStatus = namedtuple("LEDStatus", ["led_id", "led_name", "color", "blinking"])


EventLog = namedtuple("EventLog", ["event_id", "time", "device_id",
                                   "description"])


Thresholds = namedtuple("Thresholds", ["over_critical", "over_warning",
                                       "under_critical", "under_warning"])


class SensorData(object):
    """
    Base class for sensor reading data produced by fbjbod.
    """

    # error condition codes
    NORMAL = 0
    OVER_WARNING = 1
    OVER_CRITICAL = 2
    UNDER_WARNING = -1
    UNDER_CRITICAL = -2

    # type-specific configurations
    UNIT = None
    PERCENT_THRESHOLDS = False
    DATA_TYPE = None
    TYPE_NAME = None

    def __init__(self, name, value, thresholds):
        """Create a SensorData object.

        @param name: name of the sensor
        @type name: string
        @param value: the sensor's value. This value is converted to DATA_TYPE
        @type value: valid input to the DATA_TYPE (string or DATA_TYPE)
        @param thresholds: the sensor's high and low thresholds
        @type thresholds: Thresholds
        """
        self.name = name
        self.value = self.data_type(value)
        self.thresholds = thresholds
        self.threshold_state = self._check_thresholds()

    def _check_thresholds(self):
        """Checks the sensor value in relation to the threshold conditions
        that are present. These are listed as class variables, which
        are internally represented as integers. "Less than" and
        "greater than" operators are used to perform the check.
        @return int (or condition code type)
        """
        if (
                self.thresholds.over_critical is not None
                and self.value > self.thresholds.over_critical):
            return self.OVER_CRITICAL
        elif (self.thresholds.over_warning is not None
              and self.value > self.thresholds.over_warning):
            return self.OVER_WARNING
        elif (self.thresholds.under_critical is not None
              and self.value < self.thresholds.under_critical):
            return self.UNDER_CRITICAL
        elif (self.thresholds.under_warning is not None
              and self.value < self.thresholds.under_warning):
            return self.UNDER_WARNING
        else:
            return self.NORMAL

    @property
    def is_critical(self):
        """
        Sensor value is at a critical threshold
        @return bool
        """
        return (self.threshold_state is self.UNDER_CRITICAL
                or self.threshold_state is self.OVER_CRITICAL)

    @property
    def is_warning(self):
        """
        Sensor value is at a warning threshold. Does not include
        critical threshold states.
        @return bool
        """
        return (self.threshold_state is self.UNDER_WARNING
                or self.treshold_state is self.OVER_WARNING)

    @property
    def is_normal(self):
        """
        Sensor value is normal.
        @return bool
        """
        return self.threshold_state is self.NORMAL

    @property
    def is_over(self):
        """
        Sensor value is over the normal threshold
        @return bool
        """
        return self.threshold_state > self.NORMAL

    @property
    def is_under(self):
        """
        Sensor value is under the normal threshold
        @return bool
        """
        return self.threshold_state < self.NORMAL

    @classproperty
    @classmethod
    def percent_thresholds(cls):
        """
        thresholds are percentage values
        @return bool
        """
        # implemented to keep these standardized
        return cls.PERCENT_THRESHOLDS

    @classproperty
    @classmethod
    def data_type(cls):
        """
        data type of values and thresholds
        @return type
        """
        if cls.DATA_TYPE is None:
            raise NotImplementedError("'DATA_TYPE' must be implemented")
        return cls.DATA_TYPE

    @classproperty
    @classmethod
    def unit(cls):
        """
        Name of the sensor's unit
        @return string
        """
        if cls.UNIT is None:
            raise NotImplementedError("'UNIT' must be implemented")
        return cls.UNIT

    @classproperty
    @classmethod
    def unit_letter(cls):
        """
        Letter used in some keys by fbjbod for the sensor type
        @return string
        """
        return cls.unit[0].upper()

    @classproperty
    @classmethod
    def type_name(cls):
        """The name of the type. This property is a wrapper to raise an error
        if not implemented in the subclass"""
        if cls.TYPE_NAME is None:
            raise NotImplementedError("'TYPE_NAME' must be implemented")
        return cls.TYPE_NAME

    @classproperty
    @classmethod
    def type_letter(cls):
        """First letter in the unit, uppercase"""
        return cls.type_name[0].upper()

    def __str__(self):
        """Display the sensor value with the unit description, e.g. '10V'"""
        return "%s%c" % (self.value, self.unit_letter)

    def _json(self):
        """Produce a json serializable dict representation.
        @return dict
        """
        return {"name": self.name,
                "value": str(self),
                "alert": (not self.is_normal)}

    def __repr__(self):
        return '<%s: %s(%s %s)>' % (self.__class__.__name__, self.name,
                                    self.value, self.unit_letter)


class TempSensorData(SensorData):
    """
    Temperature sensor data
    """
    TYPE_NAME = "temperature"
    DATA_TYPE = int
    UNIT = "celcius"


class VoltageSensorData(SensorData):
    """
    Voltage sensor data
    """
    TYPE_NAME = "voltage"
    DATA_TYPE = float
    UNIT = "volt"


class CurrentSensorData(SensorData):
    """
    Current sensor data
    """
    TYPE_NAME = "current"
    DATA_TYPE = float
    UNIT = "amphere"
    PERCENT_THRESHOLDS = True


class PowerSensorData(SensorData):
    TYPE_NAME = "power"
    DATA_TYPE = int
    UNIT = "watt"


class EnclosureHdd(namedtuple("EnclosureHdd", ["enclosure",
                                               "slot_id",
                                               "physical_id",
                                               "sas_address",
                                               "power_status",
                                               "fault_status",
                                               "key_name"])):

    @classmethod
    def from_fbjbod_dict(cls, enclosure, hdd):
        return cls(enclosure, int(hdd["slot"]), int(hdd["phy"]),
                   hdd["sas_addr"], hdd["status"], hdd["fault"],
                   hdd["name"])

    @property
    def is_on(self):
        return self.power_status == "Power On"

    @property
    def has_fault(self):
        return self.fault_status != "None"

    def _find_drive_in_results(self, results):
        enc = results.get(self.enclosure.sg_device)
        if not enc:
            return None
        hdd = enc.get(self.key_name)
        if not hdd:
            return None
        return self.from_fbjbod_dict(self.enclosure, hdd)

    def _set_power(self, turn_on, attempts=1):
        power_str = "on" if turn_on else "off"
        res = self.enclosure._run_fbjbod("hdd", "--hdd-" + power_str,
                                         str(self.slot_id))
        d = self._find_drive_in_results(res)
        if not d:
            return False
        if d.power_status == turn_on:
            return True
        elif attempts == 1:
            return False
        else:
            time.sleep(POWER_SET_WAIT)
            return d._set_power(turn_on, attempts=(attempts - 1))

    def _set_fault(self, turn_on, attempts=1):
        fault_str = "on" if turn_on else "off"
        res = self.enclosure._run_fbjbod("hdd", "--fault-" + fault_str,
                                         str(self.slot_id))
        d = self._find_drive_in_results(res)
        if not d:
            return False
        if d.fault_status == turn_on:
            return True
        elif attempts == 1:
            return False
        else:
            time.sleep(FAULT_SET_WAIT)
            return d._set_fault(turn_on, attempts=(attempts - 1))

    def power_on(self, attempts=1):
        return self._set_power(True, attempts)

    def power_off(self, attempts=1):
        return self._set_power(False, attempts)

    def fault_enabled(self, attempts=1):
        return self._set_fault(True, attempts)

    def fault_dsiabled(self, attempts=1):
        return self._set_fault(False, attempts)

    def event_log(self):
        return (event for event in self.enclosure.event_log()
                if event.device_id == self.physical_id)

    @property
    def temperature_sensor(self):
        """Attempt to find the temperature sensor reading for the given drive
        off the enclosure
        @return TempSensorData or None
        """
        for sensor in self.enclosure.hdd_sensors:
            sensor_device_id = sensor.name.rsplit(" ", 1).pop()
            if not sensor_device_id.isdigit():
                log.debug(("Skipping believed HDD SMART temp probe with "
                           "unparsable name '%s'"),
                          sensor.name)
                continue
            if int(sensor_device_id) == self.slot_id:
                return sensor
        else:
            return None

    def _json(self):
        obj = self._asdict()
        obj["enclosure"] = str(self.enclosure)
        obj["power_status"] = self.is_on
        obj["fault_status"] = self.has_fault
        return obj


def run_fbjbod(*cmd_args):
    """Run fbjbod with supplied arguments (overloaded). The '--json'
    argument is always added, and the output parsed as json. Any
    errors encountered running fbjbod are raised as FbjbodRunError.
    OrderedDicts are used in place of dicts in order to preserve the order from
    fbjbod.
    @return OrderedDict
    """
    cmd = ["fbjbod"]
    cmd.extend(cmd_args)
    cmd.append("--json")
    log.debug("Running fbjbod command: '%s'", " ".join(cmd))
    try:
        stdout = subprocess.check_output(cmd)
    except subprocess.CalledProcessError as e:
        raise FbjbodRunError(cmd, e.returncode, "fbjbod failed to execute")
    except OSError as e:
        raise FbjbodRunError(cmd, 127, "fbjbod was not found")
    return json.loads(stdout, object_pairs_hook=OrderedDict)


def fbjbod_version():
    return run_fbjbod("version").get("Version")


class Enclosure(object):

    def __init__(self, sg_device, bsg_device=None, device_name=None):
        """Initialize the enclosure object. Some arguments are not required
        in order to leave the device path as the only key, but the
        `list_enclosures` class method adds some additional keys.

        @param sg_device: the 'sg' device path for the enclosure, e.g.
          '/dev/sg1'
        @type sg_device: string
        @param bsg_device: the bsg device path, e.g. '/dev/bsg/6:0:11:0'
        @type bsg_device: string or None
        @param device_name: The name of the enclosure, e.g. 'knox'
        @type device_name: string or None
        """
        self.sg_device = sg_device
        self._bsg_device = bsg_device
        self._device_name = device_name

        self._info = None
        self._asset_tags = None
        self._sensor_data = None

    @staticmethod
    def _list_enclosures():
        return run_fbjbod("list")

    def _set_enclosure_listing_data(self):
        """
        Normally enclosures should be created off fbjbod listing data,
        but in the event that the Enclosure object is created with just a
        sg_device, allow for the bsg_device and device_name to be
        accessible
        """
        enclosure = self._list_enclosures()[self.sg_device]
        self._bsg_device = enclosure["bsg_device"]
        self._device_name = enclosure["name"]

    @property
    def bsg_device(self):
        if self._bsg_device is None:
            self._set_enclosure_listing_data()
        return self._bsg_device

    @property
    def device_name(self):
        """
        Name of the enclosure, e.g. 'Knox'
        """
        if self._device_name is None:
            self._set_enclosure_listing_data()
        return self._device_name

    def _run_fbjbod(self, *cmd):
        """Runs fbjbod with the sg device included"""
        cmd = list(cmd)
        cmd.append(self.sg_device)
        return run_fbjbod(*cmd)

    def _gpio_status(self):
        """generator for GPIO status fields. Produces GPIOStatus tuples."""
        status = self._run_fbjbod("gpio")
        for status_dict in status.viewvalues():
            yield GPIOStatus(field_name=status_dict["name"].strip(),
                             description=status_dict["status"].strip(),
                             status=bool(int(status_dict["val"])))

    def gpio_status(self):
        return {s.field_name: s for s in self._gpio_status()}

    def info(self, reload=False):
        """Expander info, such as the firmware version, asset serials, and so
        on. No parsing is done on what fbjbod produces, but results
        are cached.

        @param reload: force rerunning the info command on fbjbod
        @type reload: bool
        @return dict
        """
        if self._info is None or not reload:
            self._info = self._run_fbjbod("info")
        return self._info

    def asset_tags(self, reload=False):
        """Expander tags, such as serial information. Can be set by
        `update_asset_tag`.

        @param reload: force rerunning the tags command on fbjbod
        @type reload: bool
        @return dict

        """
        if self._asset_tags is None or not reload:
            self._asset_tags = self._run_fbjbod("tags")
        return self._asset_tags

    def update_asset_tag(self, tag, value):
        """Update an asset tag. The cached tags are updated as well

        @param tag: the numeric tag id or tag name to update
        @type type: string or int
        @param value: the value to set on the given tag
        @type value: string
        """
        self._asset_tags = self._run_fbjbod("tag", "--id", str(tag),
                                            "--tag", str(value))

    def power_cycle(self):
        """
        Power cycle the expander
        @return bool
        """
        try:
            self._run_fbjbod("power_cycle")
        except FbjbodRunError:
            log.exception("failed to power cycle enclosure %s", self.sg_device)
            return False
        return True

    def clear_error_counters(self):
        """
        Clear physical error counters
        @return bool
        """
        try:
            self._run_fbjbod("phyerr", "--clear")
        except FbjbodRunError:
            log.exception("Failed to clear physical error counters on %s",
                          self.sg_device)
            return False
        return True

    def error_counters(self):
        """Get the physical error counters. Generator of
        `PhysicalErrorCounters` objects."""
        enc_error_counters = self._run_fbjbod("phyerr")
        for buffer_id, counters in enc_error_counters.viewitems():
            yield PhysicalErrorCounters(int(buffer_id.split("_", 1)[1]),
                                        int(counters["Invalid Dword"]),
                                        int(counters["Running Disparity"]),
                                        int(counters["Loss of Dword Sync"]),
                                        int(counters["Phy Reset Problem"]))

    def event_status(self):
        """Produce a dictions of event names to a bool indicating if it is
        active or not"""
        events = self._run_fbjbod("event", "--status")
        return {event_name.strip(): bool(int(event_data["value"]))
                for event_name, event_data in events.viewitems()}

    def _fan_status(self):
        """Generator of Fan tuples from the current fan status"""
        fans = self._run_fbjbod("fan")
        for fan in fans.viewvalues():
            yield Fan(fan["name"].strip(), int(fan["rpm"]))

    def fan_status(self):
        """
        Dictionary of fan statuses. Key is the fan name, and the value the
        current RPMs.
        @return dict
        """
        return {fan.name: fan.rpms for fan in self._fan_status()}

    def _led_status(self):
        """Generator of `LEDS,tatus`"""
        leds = self._run_fbjbod("led")
        for led in leds.viewvalues():
            status = led["status"]
            blinking = status.startswith("Blink")
            if status == "Off":
                color = None
            elif blinking:
                color = status.split(None, 1)[1]
            else:
                color = status
            yield LEDStatus(int(led["id"]), led["name"].strip(),
                            color, blinking)

    def led_status(self):
        return {led.name: led for led in self._led_status}

    @staticmethod
    def _extract_thresholds(sensor_data, sensor_type):
        """Build a `Thresholds` object based on what threshold keys can be
        found in the sensor data, assuming a pattern based off a
        letter from the sensor unit. Keys not found are represented as
        None. Otherwise, values are converted to the sensor type's
        data value.

        @param sensor_data: key vals returned for a sensor by fbjbod
        @type sensor_data: dict
        @param sensor_type: the type class to use with the sensor_data
        @type sensor_type: sub-types of SensorData
        @return Thresholds tuple
        """
        return Thresholds(
            *[(sensor_type.data_type(
                val.rstrip(" %")
                if sensor_type.percent_thresholds else val)
               if val is not None else None)
              for val in
              (sensor_data.get(key_template % sensor_type.type_letter)
               for key_template in
               ("O%c Critical", "O%c Warning", "U%c Critical", "U%c Warning"))]
        )

    SENSOR_TYPES = {
        "temperature": TempSensorData,
        "voltage": VoltageSensorData,
        "current": CurrentSensorData,
        "power": PowerSensorData,
    }

    def _sensors(self):
        """Parse through the sensor readings, and build a dictionary keyed by
        sensor type. The parsing here in generic, and data from the
        appropriate SensorDate class per type includes some parsing details.

        @return dict"""
        sensor_readings = {}
        sensor_data = self._run_fbjbod("sensor", "--thresholds")
        power = sensor_data.pop("Power", None)
        if power is not None:
            power_reading = PowerSensorData("Power", power.split(None, 1)[0],
                                            Thresholds(None, None, None, None))
            sensor_readings["power"] = [power_reading]
        for name, vals in sensor_data.viewitems():
            name = name.strip()
            # fbjbod uses a value of 0 to mean that the driver is not installed
            value = vals.get("value", "0")
            if value == "0":
                log.info("Sensor '%s' on enclosure %s is unconfigured",
                         name, self.sg_device)
                continue
            sensor_type_name = vals.get("type")
            if not sensor_type_name:
                log.error(("Sensor '%s' on enclosure %s does not have a type. "
                           "Unable to parse."),
                          name, self.sg_device)
                continue
            elif sensor_type_name not in self.SENSOR_TYPES:
                log.error(("Sensor '%s' on enclosure %s has unknown sensor "
                           "type '%s' Unable to parse."),
                          name, self.sg_device, sensor_type_name)
                continue
            sensor_type = self.SENSOR_TYPES[sensor_type_name]

            thresholds = self._extract_thresholds(vals, sensor_type)
            sensor_reading = sensor_type(name, value, thresholds)
            sensor_readings.setdefault(sensor_type_name,
                                       []).append(sensor_reading)
        return sensor_readings

    def sensors(self, max_age=60):
        """
        Wrapper for `_sensors` to cache results for a limited period of time.

        @param max_age: maximum age of sensor readings in seconds
        @type max_age: number
        @return dict
        """
        cur_time = time.time()
        if (
                self._sensor_data is None or
                (cur_time - max_age) > self._sensor_data[0]):
            self._sensor_data = (cur_time, self._sensors())
        return self._sensor_data[1]

    def hdd_sensors(self):
        """Generator for hard drive temperature sensors"""
        return (sensor for sensor
                in self.sensors().get("temperature", ())
                if sensor.name.startswith("HDD"))

    def set_fan_control(self, pwm=None, auto=False):
        """Set the fan control either by giving a pwm value or setting the fan
        to automatic bios-based control.
        @param pwm: a fan pwm value to set
        @type pwm: int or None
        @param auto: set the fan to use bios-based automatic control
        @type auto: bool
        """
        if pwm is not None and auto:
            raise ValueError("A pwm value was given and the fan set to auto")
        elif pwm is not None:
            self._run_fbjbod("fan", "--pwm", str(pwm))
        elif auto:
            self._run_fbjbod("fan", "--auto")
        else:
            raise ValueError("a fan control argument is needed")

    def config_fan_profile(self, fan_profile_id):
        """Set the fan profile
        @param fan_profile_id: the id of the fan profile to set
        @type fan_profile_id: string or int
        """
        self._run_fbjbod("config", "--fan-profile", str(fan_profile_id))

    def config_hdd_temp_read_interval(self, interval_mins):
        self._run_fbjbod("config", "--hdd-temp-int", str(interval_mins))

    def config_power_window(self, window_secs):
        self._run_fbjbod("config", "--power-win", str(window_secs))

    def event_log(self):
        """generator of EventLog tuples based on the stream of event logs. The
        ordering is unchanged from fbjbod."""
        events = self._run_fbjbod("event", "--log")
        for event in events.viewvalues():
            yield EventLog(int(event["id"]), int(event["timestamp"]),
                           int(event["id"]),
                           event["description"].strip())

    def hdds(self):
        """Generator of EnclosureHdd objects for drives on this enclosure"""
        hdd_list = self._run_fbjbod("hdd", "--all")
        if self.sg_device not in hdd_list:
            raise RuntimeError(("fbjbod did not report any "
                                "data for device %s") % self.sg_device)
        for hdd in hdd_list[self.sg_device].viewvalues():
            yield EnclosureHdd.from_fbjbod_dict(self, hdd)

    @classmethod
    def enclosures(cls):
        """Generator Enclosure objects based on the list of enclosures that
        fbjbod finds. The device name and and bsg_device are set here as
        well."""
        for enclosure in cls._list_enclosures().viewvalues():
            yield cls(sg_device=enclosure["sg_device"],
                      bsg_device=enclosure.get("bsg_device"),
                      device_name=enclosure.get("name"))

    def __repr__(self):
        return '<Enclosure(%s)>' % self.sg_device

    def __str__(self):
        return self.sg_device
